import json
from importlib import import_module
from typing import Generator, List

from pydbbackups.cli.config import BACKUPS_DATA_DIR, BACKUPS_DIR
from pydbbackups import Backup
from pydbbackups.cli.models import BackupFile, BackupData
from datetime import datetime

databases = {
    'postgres': 'Postgres',
    'mongodb': 'MongoDB'
}


def get_backups_data() -> List[BackupData]:
    b_data = BACKUPS_DATA_DIR.read_bytes()
    data = [BackupData(**v) for v in json.loads(b_data)]
    return data


def get_backups_files() -> Generator[BackupFile, None, None]:
    dir = BACKUPS_DIR.iterdir()

    for f in dir:
        [name, date] = f.name.split("___")
        yield BackupFile(name, date, f)


def get_backup_class(name: str) -> Backup:
    m = import_module(f"pydbbackups.backups.{name}")
    return getattr(m, databases[name])


def save_backup_file(meta: BackupFile, content: bytes | str) -> BackupFile:
    if "___" in meta.name:
        raise ValueError("the backup name should not include <___>")
    path = BACKUPS_DIR.joinpath(f"{meta.name}___{meta.date}")
    path.touch()
    if content is bytes:
        path.write_bytes(content)
    else:
        path.write_text(content)

    return BackupFile(meta.name, meta.date, path)


class DTEncoder(json.JSONEncoder):
    """ Datetime encoder """

    def default(self, obj):
        if isinstance(obj, datetime):
            return str(obj)
        return json.JSONEncoder.default(self, obj)



from .base import Backup
import subprocess


class MongoDB(Backup):
    CMDS_TO_CHECK = [('mongodump', '--version')]

    def dump(self):
        if self.uri:
            args = [
                '-uri', self.uri
            ]
        else:
            args = [
                '--out=mongodb_dump',
                '--username', self.username,
                '--password', self.password,
                '--host', self.host,
                '--port', f"{self.port}",
            ]

        if self.compress:
            args.append('--gzip')

        p = subprocess.Popen([
            "mongodump",
            *args
        ])

        status_code = p.wait()
        if status_code > 0:
            raise Exception(p.stderr.read().decode('utf-8'))

        return p.stdout



import click
import getpass

from rich.table import Table
from rich.console import Console
from pydbbackups.cli.config import build_config
from pydbbackups.cli.services import BackupsService


@click.group()
def app():
    """ Awesome APP """


@app.command(name='list')
def get_backups():
    table = Table()

    table.add_column('Nombre')
    table.add_column('Base de datos')
    table.add_column('Fecha')

    for bfile, bdata in BackupsService.list_backups():
        table.add_row(bfile.name, bdata.database_name, bfile.date)

    console = Console()
    console.print(table)


@app.command(name='run')
@click.option('--name', required=True)
@click.option('--database-type', required=True, type=click.Choice(['postgres', 'mongodb']))
@click.option('--host', required=True)
@click.option('--database', required=True)
@click.option('--username', default=None)
@click.option('--password', default=None)
@click.option('--without-password', default=False)
@click.option('--port', default=None)
@click.option('--compress', default=None)
def make_backup(name: str, database_type, host, database, username, password, without_password, port, compress):
    name = name.replace('-', '_')

    if without_password is False and password is None:
        password = getpass.getpass('Password: ')

    if without_password is True:
        password = None

    service = BackupsService.build(
        database_type,
        name=name,
        host=host,
        database=database,
        username=username,
        password=password,
        port=port,
        compress=compress)

    service.dump(name)


def main():
    build_config()
    app()


if __name__ == '__main__':
    main()



from .base import Backup
import subprocess
import gzip
import os
from io import BytesIO


class Postgres(Backup):

    CMDS_TO_CHECK = [('pg_dump', '--version')]

    def dump(self) -> BytesIO:
        # host = f"-h {self.host}"
        # username = f"-U {self.username}"
        # database = self.database
        # password = f'PGPASSWORD="{self.password}"' if self.password else None

        # cmd = f"{password + ' ' if password else ''} pg_dump {host} {username} {database} {'-w' if password else '--no-password'}"

        env = os.environ.copy()
        if self.password:
            env['PGPASSWORD'] = self.password

        p = subprocess.Popen([
            'pg_dump',
            '-h',
            self.host,
            '-U',
            self.username,
            self.database,
            '-w' if self.password else '--no-password'
        ], env=env, stdout=subprocess.PIPE)

        code = p.wait()
        if code > 0:
            print(p.stderr)
            raise Exception(p.stderr.read().decode('utf-8'))

        if self.compress:
            compressed = gzip.compress(p.stdout, compresslevel=9)
            return BytesIO(compressed)

        return BytesIO(p.stdout.read())



from datetime import datetime
from dataclasses import dataclass, field
from collections import namedtuple


@dataclass
class BackupData:
    id: int
    name: str
    database_name: str
    created_at: datetime = field(default_factory=lambda: datetime.now())


BackupFile = namedtuple('BackupFile', ['name', 'date', 'file'])



from pydbbackups.cli.utils import get_backup_class, save_backup_file, get_backups_data, get_backups_files, DTEncoder
from pydbbackups.cli.config import BACKUPS_DATA_DIR
from pydbbackups.cli.models import BackupData, BackupFile

from pydbbackups import Backup
from datetime import datetime
from dataclasses import asdict
import json


class BackupsService:
    backup_cls: Backup
    db_type: str

    def __init__(self, backup_cls: Backup = None) -> None:
        self.backup_cls = backup_cls

    @staticmethod
    def build(db_type: str, **kwargs):
        backup_cls = get_backup_class(db_type)(**kwargs)
        cls = BackupsService(backup_cls)
        cls.db_type = db_type

        return cls

    @staticmethod
    def list_backups():
        data = get_backups_data()
        backups = get_backups_files()

        for bfile in backups:
            bdata = next((obj for obj in data if obj.name == bfile.name), None)
            if not bdata:
                continue

            yield bfile, bdata

    def dump(self, name: str):
        output = self.backup_cls.dump()
        if not output:
            return

        created_at = datetime.now()
        meta = save_backup_file(BackupFile(
            name, created_at, ""), output.read().decode('utf-8'))
        print(meta.file, 'Created !')

        data = get_backups_data()
        data.append(BackupData(**{
            "id": len(data),
            "name": name,
            "database_name": self.db_type,
            "created_at": created_at
        }))

        serialized_data = [asdict(v) for v in data]
        BACKUPS_DATA_DIR.write_text(json.dumps(serialized_data, cls=DTEncoder))

    def restore(self):
        self.backup_cls.restore()



"""
PyDBBackups

example:

```.py
from pydbbackups import Postgres

cls = Postgres(
    name="MyDB",
    compress=False,
    database="test",
    host="localhost",
    port=5432,
    username="postgres",
    password="postgres"
)

# In some cases dump method, return None
# Like MongoDB

output = cls.dump() # Return BytesIO
print(output.read().decode('utf-8'))
```
"""

from .backups import *



from dataclasses import dataclass
from typing import Optional, Sequence, Tuple
from io import BytesIO
from pathlib import Path
from pydbbackups.errors import MethodNotImplemented, CommandNotFound
import subprocess


@dataclass
class Backup:
    host: str
    username: str
    database: str

    uri: Optional[str] = None
    name: Optional[str] = None
    password: Optional[str] = None
    port: Optional[int] = None

    compress: Optional[bool] = False

    CMDS_TO_CHECK: Optional[Sequence[Tuple[str, Optional[str]] | str]] = None

    @classmethod
    def exist_cmds(cls):
        if not cls.CMDS_TO_CHECK:
            print(
                f"Warning : The {cls.__name__} class doesn't have commands to check")
            return

        for cmd in cls.CMDS_TO_CHECK:
            args = []
            if isinstance(cmd, tuple):
                command = cmd[0]
                if len(cmd) > 1:
                    args = cmd[1]
            else:
                command = cmd

            try:
                subprocess.Popen([
                    command,
                    *args
                ], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            except FileNotFoundError:
                raise CommandNotFound(command)

    def dump(self) -> BytesIO:
        raise MethodNotImplemented('Dump')

    def restore(self, file_path: Path) -> BytesIO:
        raise MethodNotImplemented('Restore')

    def __post_init__(self):
        self.exist_cmds()
